
clear all

%========================================= Notes ===================================================

% This script performs covariance between pairs of spectra with dimensions I by X by J and 
% K by X by L. The result is a 4D spectrum with dimensions I, J, K and L.

% Additionally, several pairs of spectra can be covaried at once and then multiplied together to
% further enhance the results. In these cases, each pair of spectra must have the same values
% of I, J, K and L as all other pairs.

% For a sequential HN,HN assignment spectrum I and K are 1H dimenions; X is a CA, CB or CO 
% dimension; and J and L are 15N dimensions.

% The 4D is processed by covarying each I by X plane with each K by X plane, resulting in an
% I by K plane. The X dimension is subsumed. We loop over the dimensions J and L.
% As a result, the dimension order of the 4D is IKLJ. Projections of this 4D are written to the 
% disk without any transpose. It is the responsibility of the user to transpose as needed.

%===================================================================================================


%===================================================================================================
%                          Users should only need to edit this section 
%===================================================================================================

% A data structure is used to store all of the relevant information for a spectrum pair.

% The add_spectrum_pair function extracts the relevant information from the nmrPipe headers and
% returns a data structure.

% The add_spectrum_pair function takes two or three arguments. The first two are format strings
% used to locate and load the pairs of nmrPipe spectra. Be sure to be consistent in ordering here
% if  you plan to multiply the covariance results of several pairs.

% If no third argument is supplied, the add_spectrum_pair function creates a new data structure
% for the spectrum pair and returns it. Alternatively, if you supply a previously created data
% structure as the third argument, the function will add the new spectrum pair to the existing
% data structure and return it.

% If the data structure contains multiple pairs of spectra, the covariance results of each pair are
% automatically multiplied together during processing.

% Be sure to use the name "structure" for the data structure (it is required in the code below)
structure = add_spectrum_pair('./HMCMCBCA_VB/test%04d.ft3', ...
                              './HNCACB_VB/test%03d.ft3');

structure = add_spectrum_pair('./HMCMCBCA_VA/test%04d.ft3', ...
                              './HNCA_VA/test%03d.ft3', structure);

% structure = add_spectrum_pair('./data/HNCACB/ft3_covar/test%04d.ft3', ...
%                               './data/HNCOCACB/ft3_covar/test%04d.ft3', structure);

% 3D projection output formats and dimension labels
% Be sure to match the order of dimension labels to the data.
% You must provide exactly 3 labels. Labels may be no more than 4 characters.
IKJ_format = './V_HmeHC/test%04d.ft3';
IKJ_label = {'Hme', 'H', 'C'};

ILJ_format = './V_HmeNC/test%04d.ft3';
ILJ_label = {'Hme', 'N', 'C'};

IKL_format = './V_HmeHN/test%04d.ft3';
IKL_label = {'Hme', 'H', 'N'};

KLJ_format = './V_HNC/test%04d.ft3';
KLJ_label = {'H', 'N', 'C'};

%4D
IKLJ_label = {'Hme', 'H','N', 'C'};

%===================================================================================================



% Small variables used repeatedly
I = structure.I;
J = structure.J;
K = structure.K;
L = structure.L;
np = structure.num_pairs;



%---------------------------- Create output headers from input headers -----------------------------

% The covariance toolbox doesn't take care to set the FDFILECOUNT variable.
% This may be a problem at some point

template_IJ = structure.headers{1,1};
template_KL = structure.headers{2,1};

% Create output headers from templates, correct dimension sizes and add dimension labels
IKJ_header = [template_IJ(:,1) template_KL(:,1) template_IJ(:,3:4)];
IKJ_header(26,1:3) = [I K J];
IKJ_header(12,1:3) = cellfun( @(label) typecast([uint8(label) zeros(1, 4-length(label))], ...
                                                'single'), IKJ_label);

IKL_header = [template_IJ(:,1) template_KL(:,1) template_KL(:,3:4)];
IKL_header(26,1:3) = [I K L];
IKL_header(12,1:3) = cellfun( @(label) typecast([uint8(label) zeros(1, 4-length(label))], ...
                                                'single'), IKL_label);

ILJ_header = [template_IJ(:,1) template_KL(:,3) template_IJ(:,3:4)];
ILJ_header(26,1:3) = [I L J];
ILJ_header(12,1:3) = cellfun( @(label) typecast([uint8(label) zeros(1, 4-length(label))], ...
                                                'single'), ILJ_label);

KLJ_header = [template_KL(:,1) template_KL(:,3) template_IJ(:,3:4)];
KLJ_header(26,1:3) = [K L J];
KLJ_header(12,1:3) = cellfun( @(label) typecast([uint8(label) zeros(1, 4-length(label))], ...
                                                'single'), KLJ_label);
                                            
%4D

IKLJ_header=[template_IJ(:,1) template_KL(:,1) template_KL(:,3) template_IJ(:,3)];
IKLJ_header(26,:)=[I K L J];
IKLJ_header(12,:)=cellfun( @(label) typecast([uint8(label) zeros(1, 4-length(label))], ...
                                                'single'), IKLJ_label);

%---------------------------------------------------------------------------------------------------


%------------------- Calculate planes of covariance 4D and create each projection ------------------

% We will never actually write the 4D to the disk.
% Most projections can be produced a plane at a time and written to the disk.
IKJ = zeros(I,K);
ILJ = zeros(I,L);
IKL = zeros(I,K,L); % Must store entire 3D. Summing over outer loop
KLJ = zeros(K,L);

% Initialize cell array for storing raw data planes.
raw_planes = cell(2, np);
% Not sure if initializing arrays helps here. If not, I can remove subsumed from the structure
for m = 1:np
    raw_planes(:,m) = { zeros(I, structure.subsumed(m)); ...
                        zeros(K, structure.subsumed(m)) };
end

% Initialize three dimensional array for storing covariance results
covar_planes = zeros(I, K, np);

wbar = waitbar(0,'Processing...');
tic

for j = 1:J
    j_a=num2str(j,'%03d');
    % Read jth plane from the I by X by J spectrum, 
    % take the derivative along the subsumed X dimension and store
    for m = 1:np
        fileID = fopen( sprintf(structure.formats{1,m}, j), 'r');
        fseek(fileID, 2048, 'bof'); % Skip header
        raw_planes{1,m} = gradient( fread(fileID, [I, Inf], 'float32') );
        fclose(fileID);
    end
    
    % Reset projections that do not sum over J
    IKJ = zeros(I,K);
    ILJ = zeros(I,L);
    KLJ = zeros(K,L);
    
    for l = 1:L
       l_z=num2str(l,'%03d');     
        % Read lth plane from the K by X by L spectrum, 
        % take the derivative along the subsumed X dimension and store
        for m = 1:np
            fileID = fopen( sprintf(structure.formats{2,m}, l), 'r');
            fseek(fileID, 2048, 'bof');
            raw_planes{2,m} = gradient( fread(fileID, [K, Inf], 'float32') );
            fclose(fileID);
        end
        
        % Perform covariance and take product of covaried planes
        for m = 1:np
            covar_planes(:,:,m) = covar( raw_planes{1,m}, raw_planes{2,m}, ...
                                          'axes1', structure.headers{1,m},   ...
                                          'axes2', structure.headers{2,m} );
        end
        final_plane = prod(covar_planes, 3);
        
        % Calculate projections
        IKJ = IKJ + final_plane;
        IKL(:,:,l) = IKL(:,:,l) + final_plane;
        ILJ(:,l) = sum(final_plane, 2);
        KLJ(:,l) = sum(final_plane, 1)';
        % write 4D
        test=['test' j_a l_z '.ft4']; 
        tmp=['V_HmHNC/' test];
        write_covar2pipe(final_plane,IKLJ_header,tmp)
    end
    
    % Write out planes of projections that do not sum over the J dimension
    write_covar2pipe(IKJ, IKJ_header, sprintf(IKJ_format, j));
    write_covar2pipe(ILJ, ILJ_header, sprintf(ILJ_format, j));
    write_covar2pipe(KLJ, KLJ_header, sprintf(KLJ_format, j));
    
    waitbar(j/J);
end

% Write final projection
for l = 1:L
    write_covar2pipe(IKL(:,:,l), IKL_header, sprintf(IKL_format, l));
end

toc
close(wbar);


